# pytest-anki
#
# Copyright (C)  2019-2021 Aristotelis P. <https://glutanimate.com/>
#                and contributors (see CONTRIBUTORS file)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version, with the additions
# listed at the end of the license file that accompanied this program.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# NOTE: This program is subject to certain additional terms pursuant to
# Section 7 of the GNU Affero General Public License.  You should have
# received a copy of these additional terms immediately following the
# terms and conditions of the GNU Affero General Public License that
# accompanied this program.
#
# If not, please request a copy through one of the means of contact
# listed here: <https://glutanimate.com/contact/>.
#
# Any modifications to this file must keep this entire header intact.


from pathlib import Path
from typing import Tuple, Union

PathLike = Union[str, Path]
UnpackedAddon = Tuple[PathLike, str]  # path to add-on folder, package name



# pytest-anki
#
# Copyright (C)  2019-2021 Aristotelis P. <https://glutanimate.com/>
#                and contributors (see CONTRIBUTORS file)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version, with the additions
# listed at the end of the license file that accompanied this program.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# NOTE: This program is subject to certain additional terms pursuant to
# Section 7 of the GNU Affero General Public License.  You should have
# received a copy of these additional terms immediately following the
# terms and conditions of the GNU Affero General Public License that
# accompanied this program.
#
# If not, please request a copy through one of the means of contact
# listed here: <https://glutanimate.com/contact/>.
#
# Any modifications to this file must keep this entire header intact.

import shutil
from pathlib import Path
from typing import Any, Dict, NamedTuple, Optional

from aqt.addons import AddonManager

from ._types import PathLike
from ._util import create_json


def _to_path(path: PathLike) -> Path:
    path_obj = Path(path)
    if not path_obj.exists():
        raise IOError("Provided path does not exist")
    return path_obj


def install_addon_from_package(addon_manager: AddonManager, addon_path: PathLike):
    addon_path = _to_path(addon_path)
    if addon_path.suffix != ".ankiaddon":
        raise ValueError("Provided path is not an .ankiaddon file")
    addon_manager.install(str(addon_path))


def install_addon_from_folder(
    anki_base_dir: PathLike,
    package_name: str,
    addon_path: PathLike,
):
    addon_path = _to_path(addon_path)
    anki_base_dir = _to_path(anki_base_dir)
    if not addon_path.is_dir() or not anki_base_dir.is_dir():
        raise ValueError("Provided path is not a folder")
    if not package_name:
        raise ValueError("Package name must not be empty")
    destination_path = anki_base_dir / "addons21" / package_name
    shutil.copytree(src=addon_path, dst=destination_path, dirs_exist_ok=True)


class ConfigPaths(NamedTuple):
    default_config: Optional[Path]
    user_config: Optional[Path]


def create_addon_config(
    anki_base_dir: PathLike,
    package_name: str,
    default_config: Optional[Dict[str, Any]] = None,
    user_config: Optional[Dict[str, Any]] = None,
) -> ConfigPaths:
    addon_path = Path(anki_base_dir) / "addons21" / package_name
    addon_path.mkdir(parents=True, exist_ok=True)

    defaults_path = meta_path = None

    if default_config:
        defaults_path = addon_path / "config.json"
        create_json(defaults_path, default_config)

    if user_config:
        meta_path = addon_path / "meta.json"
        create_json(meta_path, {"config": user_config})

    return ConfigPaths(defaults_path, meta_path)



# pytest-anki
#
# Copyright (C)  2019-2021 Aristotelis P. <https://glutanimate.com/>
#                and contributors (see CONTRIBUTORS file)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version, with the additions
# listed at the end of the license file that accompanied this program.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# NOTE: This program is subject to certain additional terms pursuant to
# Section 7 of the GNU Affero General Public License.  You should have
# received a copy of these additional terms immediately following the
# terms and conditions of the GNU Affero General Public License that
# accompanied this program.
#
# If not, please request a copy through one of the means of contact
# listed here: <https://glutanimate.com/contact/>.
#
# Any modifications to this file must keep this entire header intact.

from typing import TYPE_CHECKING, Any, Dict, Iterator, Optional

import pytest

if TYPE_CHECKING:
    from pytest import FixtureRequest
    from pytestqt.qtbot import QtBot

from ._launch import anki_running
from ._session import AnkiSession


@pytest.fixture
def anki_session(request: "FixtureRequest", qtbot: "QtBot") -> Iterator[AnkiSession]:
    """Fixture that instantiates Anki, yielding an AnkiSession object

    All keyword arguments below may be passed to the fixture by using indirect
    parametrization.

    E.g., to specify a custom profile name you would decorate your test method with:

    > @pytest.mark.parametrize("anki_session", [dict(profile_name="foo")],
                               indirect=True)

    Keyword Arguments:
        base_path {str} -- Path to write Anki base folder to
            (default: system-wide temporary directory)

        base_name {str} -- Base folder name (default: {"anki_base"})

        profile_name {str} -- User profile name (default: {"User 1"})

        lang {str} -- Language to use for the user profile (default: {"en_US"})

        load_profile {bool} -- Whether to preload Anki user profile (with collection)
            (default: {False})

        preset_anki_state {Optional[pytest_anki.AnkiStateUpdate]}:
            Allows pre-configuring Anki object state, as described by a PresetAnkiState
            dataclass. This includes the three main configuration storages used by
            add-ons, mw.col.conf (colconf_storage), mw.pm.profile (profile_storage),
            and mw.pm.meta (meta_storage).

            The provided data is applied on top of the existing data in each case
            (i.e. in the same way as dict.update(new_data) would).

            State specified in this manner is guaranteed to be pre-configured ahead of
            add-on load time (in the case of meta_storage), or ahead of
            gui_hooks.profile_did_open fire time (in the case of colconf_storage and
            profile_storage).

            Please note that, in the case of colconf_storage and profile_storage, the
            caller is responsible for either passing 'load_profile=True', or manually
            loading the profile at a later stage.

        packed_addons {Optional[List[PathLike]]}: List of paths to .ankiaddon-packaged
            add-ons that should be installed ahead of starting Anki

        unpacked_addons {Optional[List[Tuple[str, PathLike]]]}:
            List of unpacked add-ons that should be installed ahead of starting Anki.
            Add-ons need to be specified as tuple of the add-on package name under which
            to install the add-on, and the path to the source folder (the package
            folder containing the add-on __init__.py)

        addon_configs {Optional[List[Tuple[str, Dict[str, Any]]]]}:
            List of add-on package names and config values to set the user configuration
            for the specified add-on to. Useful for simulating specific config set-ups.
            Each list member needs to be specified as a tuple of add-on package name
            and dictionary of user configuration values to set.

        web_debugging_port {Optional[int]}:
            If specified, launches Anki with QTWEBENGINE_REMOTE_DEBUGGING set, allowing
            you to remotely debug Qt web engine views.

    """

    indirect_parameters: Optional[Dict[str, Any]] = getattr(request, "param", None)

    with anki_running(qtbot=qtbot) if not indirect_parameters else anki_running(
        qtbot=qtbot, **indirect_parameters
    ) as session:
        yield session



# pytest-anki
#
# Copyright (C)  2019-2021 Aristotelis P. <https://glutanimate.com/>
#                and contributors (see CONTRIBUTORS file)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version, with the additions
# listed at the end of the license file that accompanied this program.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# NOTE: This program is subject to certain additional terms pursuant to
# Section 7 of the GNU Affero General Public License.  You should have
# received a copy of these additional terms immediately following the
# terms and conditions of the GNU Affero General Public License that
# accompanied this program.
#
# If not, please request a copy through one of the means of contact
# listed here: <https://glutanimate.com/contact/>.
#
# Any modifications to this file must keep this entire header intact.

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

from ._errors import AnkiSessionError
from ._util import get_nested_attribute

if TYPE_CHECKING:
    from anki.collection import Collection
    from anki.config import ConfigManager
    from aqt.main import AnkiQt


class AnkiWebViewType(Enum):
    main_webview = "main webview"
    top_toolbar = "top toolbar"
    bottom_toolbar = "bottom toolbar"
    legacy_deck_stats = "deck stats"
    previewer = "previewer"
    browser_card_info = "browser card info"
    card_layout = "card layout"
    change_notetype = "changeNotetype"
    find_duplicates = "find duplicates"
    empty_cards = "empty cards"


@dataclass
class AnkiStateUpdate:

    """
    Specifies Anki object state to be pre-configured for the test session.

    This includes the three main configuration storages used by add-ons:

    - mw.col.conf (colconf_storage), available at profile load time
    - mw.pm.profile (profile_storage), available at profile load time
    - mw.pm.meta (meta_storage), available at add-on load time
    """

    colconf_storage: Optional[Dict[str, Any]] = None
    profile_storage: Optional[Dict[str, Any]] = None
    meta_storage: Optional[Dict[str, Any]] = None


class AnkiStorageObject(Enum):
    colconf_storage = "col.conf"
    profile_storage = "pm.profile"
    meta_storage = "pm.meta"


def get_collection(main_window: "AnkiQt") -> "Collection":
    if (collection := main_window.col) is None:
        raise AnkiSessionError(
            "Collection has not been loaded, yet. Please use load_profile()."
        )
    return collection


def update_anki_profile_state(
    main_window: "AnkiQt", anki_state_update: AnkiStateUpdate
):
    _update_single_anki_state(
        main_window=main_window,
        storage_object=AnkiStorageObject.profile_storage,
        anki_state_update=anki_state_update,
    )


def update_anki_meta_state(main_window: "AnkiQt", anki_state_update: AnkiStateUpdate):
    _update_single_anki_state(
        main_window=main_window,
        storage_object=AnkiStorageObject.meta_storage,
        anki_state_update=anki_state_update,
    )


def update_anki_colconf_state(
    main_window: "AnkiQt", anki_state_update: AnkiStateUpdate
):
    _update_single_anki_state(
        main_window=main_window,
        storage_object=AnkiStorageObject.colconf_storage,
        anki_state_update=anki_state_update,
    )


def _update_single_anki_state(
    main_window: "AnkiQt",
    storage_object: AnkiStorageObject,
    anki_state_update: AnkiStateUpdate,
):
    data: Optional[Dict[str, Any]] = getattr(anki_state_update, storage_object.name)
    if data:
        set_anki_object_data(
            main_window=main_window, storage_object=storage_object, data=data
        )


def update_anki_state(main_window: "AnkiQt", anki_state_update: AnkiStateUpdate):
    if anki_state_update.meta_storage:
        update_anki_meta_state(
            main_window=main_window, anki_state_update=anki_state_update
        )

    if anki_state_update.profile_storage:
        update_anki_profile_state(
            main_window=main_window, anki_state_update=anki_state_update
        )

    if anki_state_update.colconf_storage:
        update_anki_colconf_state(
            main_window=main_window, anki_state_update=anki_state_update
        )


def set_anki_object_data(
    main_window: "AnkiQt", storage_object: AnkiStorageObject, data: dict
) -> Union[Dict[str, Any], "ConfigManager"]:
    """Update the data of a specified Anki storage object

    This may be used to simulate specific Anki and/or add-on states
    during testing."""

    anki_object = get_anki_object(
        main_window=main_window, storage_object=storage_object
    )

    if storage_object == AnkiStorageObject.colconf_storage:
        # mw.col.conf dict API is deprecated in favor of ConfigManager API
        collection = get_collection(main_window=main_window)
        for key, value in data.items():
            collection.set_config(key, value)
    else:
        anki_object.update(data)  # type: ignore

    return anki_object


def get_anki_object(
    main_window: "AnkiQt", storage_object: AnkiStorageObject
) -> Union[Dict[str, Any], "ConfigManager"]:
    """Get Anki object for specified AnkiStorageObject type"""
    attribute_path = storage_object.value
    try:
        return get_nested_attribute(obj=main_window, attr=attribute_path)
    except Exception as e:
        raise AnkiSessionError(
            f"Anki storage object {storage_object.name} could not be accessed:"
            f" {str(e)}"
        )



# pytest-anki
#
# Copyright (C)  2019-2021 Aristotelis P. <https://glutanimate.com/>
#                and contributors (see CONTRIBUTORS file)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version, with the additions
# listed at the end of the license file that accompanied this program.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# NOTE: This program is subject to certain additional terms pursuant to
# Section 7 of the GNU Affero General Public License.  You should have
# received a copy of these additional terms immediately following the
# terms and conditions of the GNU Affero General Public License that
# accompanied this program.
#
# If not, please request a copy through one of the means of contact
# listed here: <https://glutanimate.com/contact/>.
#
# Any modifications to this file must keep this entire header intact.

import re
from contextlib import contextmanager
from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)

from anki.importing.apkg import AnkiPackageImporter
from PyQt5.QtCore import QThreadPool, QTimer
from PyQt5.QtWebEngineWidgets import QWebEngineProfile
from selenium import webdriver

from ._addons import ConfigPaths, create_addon_config
from ._anki import AnkiStateUpdate, AnkiWebViewType, get_collection, update_anki_state
from ._errors import AnkiSessionError
from ._qt import SignallingWorker
from ._types import PathLike

if TYPE_CHECKING:
    from anki.collection import Collection
    from aqt import AnkiApp
    from aqt.main import AnkiQt
    from pytestqt.qtbot import QtBot


class AnkiSession:
    def __init__(
        self,
        app: "AnkiApp",
        mw: "AnkiQt",
        user: str,
        base: str,
        qtbot: "QtBot",
        web_debugging_port: Optional[int] = None,
    ):
        """Anki test session object, returned by anki_session fixture.

        Contains a number of helpful properties and methods to characterize and
        interact with a running Anki test session.

        Arguments:
            app {AnkiApp} -- Anki QApplication instance
            mw {AnkiQt} -- Anki QMainWindow instance
            user {str} -- User profile name (e.g. "User 1")
            base {str} -- Path to Anki base directory
        """

        self._app = app
        self._mw = mw
        self._user = user
        self._base = base
        self._qtbot = qtbot
        self._web_debugging_port = web_debugging_port
        self._chrome_driver: Optional[webdriver.Chrome] = None

    # Key session properties ####

    @property
    def app(self) -> "AnkiApp":
        """Anki's current QApplication instance"""
        return self._app

    @property
    def mw(self) -> "AnkiQt":
        """Anki's current main window instance"""
        return self._mw

    @property
    def user(self) -> str:
        """The current user profile name (e.g. 'User 1')"""
        return self._user

    @property
    def base(self) -> str:
        """Path to Anki base directory"""
        return self._base

    # Interaction with Qt

    @property
    def qtbot(self) -> "QtBot":
        """pytest-qt QtBot fixture"""
        return self._qtbot

    # Web view debugging

    @property
    def web_debugging_port(self) -> Optional[int]:
        """Port used for remote web debugging (if set)"""
        return self._web_debugging_port

    @property
    def chromium_version(self) -> str:
        user_agent = QWebEngineProfile.defaultProfile().httpUserAgent()
        match = re.match(r".*Chrome/(.+)\s+.*", user_agent)
        if match is None:
            raise AnkiSessionError("Could not determine Chromium version")
        return match.groups()[0]

    # Collection and profiles ####

    @property
    def collection(self) -> "Collection":
        """Returns current Anki collection if loaded"""
        return get_collection(self._mw)

    def load_profile(self) -> "Collection":
        """Load Anki profile, returning user collection

        Note: In a multi-profile configuration this method will raise the profile
        selection dialog, blocking until a profile is selected via UI interaction
        """
        self._mw.setupProfile()
        if self._mw.col is None:
            raise AnkiSessionError("Could not load collection")
        return self._mw.col

    def unload_profile(self, on_profile_unloaded: Optional[Callable] = None):
        """Unload current profile, optionally running a callback when profile
        unload complete"""
        if on_profile_unloaded is None:
            on_profile_unloaded = lambda *args, **kwargs: None  # noqa: E731
        self._mw.unloadProfile(on_profile_unloaded)

    @contextmanager
    def profile_loaded(self) -> Iterator["Collection"]:
        """Context manager that takes care of loading and then tearing down
        user profile"""
        collection = self.load_profile()

        yield collection

        self.unload_profile()

    # Deck management ####

    def install_deck(self, path: PathLike) -> int:
        """Install deck from specified .apkg file, returning deck ID"""
        old_ids = set(self._get_deck_ids())

        importer = AnkiPackageImporter(col=self.collection, file=str(path))
        importer.run()

        new_ids = set(self._get_deck_ids())

        # deck IDs are strings on <=2.1.26
        deck_id = int(next(iter(new_ids - old_ids)))

        return deck_id

    def remove_deck(self, deck_id: int):
        """Remove deck as specified by provided deck ID"""
        try:  # 2.1.28+
            # Deck methods on 2.1.45 and up use a DeckId NewType derived from int.
            # This only makes a difference at type-check time, so we stick with
            # passing in an int for now.
            self.collection.decks.remove([deck_id])  # type: ignore[list-item]
        except AttributeError:  # legacy
            self.collection.decks.rem(deck_id, cardsToo=True)

    @contextmanager
    def deck_installed(self, path: PathLike) -> Iterator[int]:
        """Context manager that takes care of installing deck and then removing
        it upon context completion"""
        deck_id = self.install_deck(path=path)

        yield deck_id

        self.remove_deck(deck_id=deck_id)

    def _get_deck_ids(self) -> List[int]:
        try:  # 2.1.28+
            return [d.id for d in self.collection.decks.all_names_and_ids()]
        except AttributeError:  # legacy
            return self.collection.decks.allIds()  # type: ignore[attr-defined]

    # Add-on loading ####

    def load_addon(self, package_name: str) -> ModuleType:
        """Dynamically import an add-on as specified by its package name"""
        addon_package = __import__(package_name)
        return addon_package

    # Add-on config handling ####

    def create_addon_config(
        self,
        package_name: str,
        default_config: Optional[Dict[str, Any]] = None,
        user_config: Optional[Dict[str, Any]] = None,
    ) -> ConfigPaths:
        """Create and populate the config.json and meta.json configuration
        files for an add-on, as specified by its package name"""
        if default_config is None and user_config is None:
            raise ValueError(
                "Need to provide at least one of default_config, user_config"
            )

        return create_addon_config(
            anki_base_dir=self._base,
            package_name=package_name,
            default_config=default_config,
            user_config=user_config,
        )

    @contextmanager
    def addon_config_created(
        self,
        package_name: str,
        default_config: Optional[Dict[str, Any]] = None,
        user_config: Optional[Dict[str, Any]] = None,
    ) -> Iterator[ConfigPaths]:
        """Context manager that takes care of creating the configuration files
        for an add-on, as specified by its package name, and then deleting them
        upon context exit."""
        if default_config is None and user_config is None:
            raise ValueError(
                "Need to provide at least one of default_config, user_config"
            )

        config_paths = self.create_addon_config(
            package_name=package_name,
            default_config=default_config,
            user_config=user_config,
        )

        yield config_paths

        if config_paths.default_config and config_paths.default_config.exists():
            config_paths.default_config.unlink()

        if config_paths.user_config and config_paths.user_config.exists():
            config_paths.user_config.unlink()

    # Anki config object handling ####

    def update_anki_state(self, anki_state_update: AnkiStateUpdate):
        """Set the state of certain Anki storage objects that are frequently used by add-ons.
        This includes mw.col.conf (colconf_storage), mw.pm.profile (profile_storage),
        and mw.pm.meta (meta_storage).

        The combined state of all objects is supplied as a pytest_anki.AnkiStateUpdate
        data class.
        """
        update_anki_state(main_window=self._mw, anki_state_update=anki_state_update)

    # Synchronicity / event loop handling ####

    def run_in_thread_and_wait(
        self,
        task: Callable,
        task_args: Optional[Tuple[Any, ...]] = None,
        task_kwargs: Optional[Dict[str, Any]] = None,
    ):
        thread_pool = QThreadPool.globalInstance()
        worker = SignallingWorker(
            task=task, task_args=task_args, task_kwargs=task_kwargs
        )

        with self._qtbot.wait_signal(worker.signals.finished):
            thread_pool.start(worker)

        if exception := worker.error:
            raise exception

        return worker.result

    def set_timeout(self, task: Callable, delay: int, *args, **kwargs):
        QTimer.singleShot(delay, lambda: task(*args, **kwargs))

    # Web debugging ####

    @contextmanager
    def _allow_selenium_to_detect_anki(self) -> Iterator[None]:
        """
        Context manager that reversibly patches Anki's application name and
        version, so that it passes Selenium's logic for identifying
        supported browsers

        cf. https://forum.qt.io/topic/96202
        """
        old_application_name = self.mw.app.applicationName()
        old_application_version = self.mw.app.applicationVersion()
        self.mw.app.setApplicationName("Chrome")
        self.mw.app.setApplicationVersion(self.chromium_version)
        yield
        self.mw.app.setApplicationName(old_application_name)
        self.mw.app.setApplicationVersion(old_application_version)

    def _switch_chrome_driver_to_web_view(
        self, driver: webdriver.Chrome, web_view_title: str
    ):
        for window_handle in driver.window_handles:
            driver.switch_to.window(window_handle)
            if driver.title == web_view_title:
                break
        else:
            raise AnkiSessionError(
                f"Could not find web view with provided title '{web_view_title}'"
            )

    def run_with_chrome_driver(
        self,
        test_function: Callable[[webdriver.Chrome], Optional[bool]],
        target_web_view: Optional[Union[AnkiWebViewType, str]] = None,
    ):
        """[summary]

        Args:
            test_function (Callable[[webdriver.Chrome], Optional[bool]]): [description]
            target_web_view: Web view as identified by its title. Defaults to None.
        """
        if self._web_debugging_port is None:
            raise AnkiSessionError("Web debugging interface is not active")

        web_view_title: Optional[str]

        if isinstance(target_web_view, AnkiWebViewType):
            web_view_title = target_web_view.value
        else:
            web_view_title = target_web_view

        def test_wrapper() -> Optional[bool]:
            if not self._chrome_driver:
                options = webdriver.ChromeOptions()
                options.add_experimental_option(
                    "debuggerAddress", f"127.0.0.1:{self._web_debugging_port}"
                )
                self._chrome_driver = webdriver.Chrome(options=options)

            if web_view_title:
                self._switch_chrome_driver_to_web_view(
                    driver=self._chrome_driver, web_view_title=web_view_title
                )

            return test_function(self._chrome_driver)

        with self._allow_selenium_to_detect_anki():
            return self.run_in_thread_and_wait(test_wrapper)

    def reset_chrome_driver(self):
        if not self._chrome_driver:
            return
        self._chrome_driver.quit()
        self._chrome_driver = None



# pytest-anki
#
# Copyright (C)  2019-2021 Aristotelis P. <https://glutanimate.com/>
#                and contributors (see CONTRIBUTORS file)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version, with the additions
# listed at the end of the license file that accompanied this program.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# NOTE: This program is subject to certain additional terms pursuant to
# Section 7 of the GNU Affero General Public License.  You should have
# received a copy of these additional terms immediately following the
# terms and conditions of the GNU Affero General Public License that
# accompanied this program.
#
# If not, please request a copy through one of the means of contact
# listed here: <https://glutanimate.com/contact/>.
#
# Any modifications to this file must keep this entire header intact.


from typing import Any, Callable, Dict, Optional, Tuple

from PyQt5.QtCore import QMessageLogContext, QObject, QRunnable, QtMsgType, pyqtSignal


class QtMessageMatcher(QObject):

    match_found = pyqtSignal()

    def __init__(self, matched_phrase: str):
        super().__init__()
        self._matched_phrase = matched_phrase

    def __call__(self, type: QtMsgType, context: QMessageLogContext, msg: str):
        if self._matched_phrase in msg:
            self.match_found.emit()


class Signals(QObject):
    finished = pyqtSignal()


class SignallingWorker(QRunnable):
    def __init__(
        self,
        task: Callable,
        task_args: Optional[Tuple[Any, ...]] = None,
        task_kwargs: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.signals = Signals()
        self._task = task
        self._task_args = task_args or tuple()
        self._task_kwargs = task_kwargs or {}
        self._result: Optional[Any] = None
        self._error: Optional[Exception] = None

    def run(self):
        try:
            self._result = self._task(*self._task_args, **self._task_kwargs)
        except Exception as error:
            self._error = error
        finally:
            self.signals.finished.emit()

    @property
    def result(self) -> Optional[Any]:
        return self._result

    @property
    def error(self) -> Optional[Exception]:
        return self._error



# pytest-anki
#
# Copyright (C)  2017-2021 Ankitects Pty Ltd and contributors
# Copyright (C)  2017-2019 Michal Krassowski <https://github.com/krassowski>
# Copyright (C)  2019-2021 Aristotelis P. <https://glutanimate.com/>
#                and contributors (see CONTRIBUTORS file)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version, with the additions
# listed at the end of the license file that accompanied this program.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# NOTE: This program is subject to certain additional terms pursuant to
# Section 7 of the GNU Affero General Public License.  You should have
# received a copy of these additional terms immediately following the
# terms and conditions of the GNU Affero General Public License that
# accompanied this program.
#
# If not, please request a copy through one of the means of contact
# listed here: <https://glutanimate.com/contact/>.
#
# Any modifications to this file must keep this entire header intact.


import uuid
from argparse import Namespace
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterator, List, Optional, Tuple
from unittest.mock import Mock

import aqt
from aqt.main import AnkiQt
from aqt.mediasync import MediaSyncer
from aqt.taskman import TaskManager
from PyQt5.QtWidgets import QMainWindow

if TYPE_CHECKING:
    from anki._backend import RustBackend
    from anki.collection import Collection
    from aqt.profiles import ProfileManager as ProfileManagerType

from ._addons import (
    create_addon_config,
    install_addon_from_folder,
    install_addon_from_package,
)
from ._anki import AnkiStateUpdate, update_anki_meta_state
from ._types import PathLike

PostUISetupCallbackType = Callable[[AnkiQt], None]


def post_ui_setup_callback_factory(
    anki_base_dir: PathLike,
    packed_addons: Optional[List[PathLike]] = None,
    unpacked_addons: Optional[List[Tuple[str, PathLike]]] = None,
    addon_configs: Optional[List[Tuple[str, Dict[str, Any]]]] = None,
    preset_anki_state: Optional[AnkiStateUpdate] = None,
):
    def post_ui_setup_callback(main_window: AnkiQt):
        """Initialize add-on manager, install add-ons, load add-ons"""
        main_window.addonManager = aqt.addons.AddonManager(main_window)

        if packed_addons:
            for packed_addon in packed_addons:
                install_addon_from_package(
                    addon_manager=main_window.addonManager, addon_path=packed_addon
                )

        if unpacked_addons:
            for package_name, addon_path in unpacked_addons:
                install_addon_from_folder(
                    anki_base_dir=anki_base_dir,
                    package_name=package_name,
                    addon_path=addon_path,
                )

        if addon_configs:
            for package_name, config_values in addon_configs:
                create_addon_config(
                    anki_base_dir=anki_base_dir,
                    package_name=package_name,
                    user_config=config_values,
                )

        if preset_anki_state and preset_anki_state.meta_storage:
            update_anki_meta_state(
                main_window=main_window, anki_state_update=preset_anki_state
            )

        main_window.addonManager.loadAddons()

    return post_ui_setup_callback


def custom_init_factory(post_ui_setup_callback: PostUISetupCallbackType):
    def custom_init(
        main_window: AnkiQt,
        app: aqt.AnkiApp,
        profileManager: "ProfileManagerType",
        backend: "RustBackend",
        opts: Namespace,
        args: List[Any],
        **kwargs,
    ):
        import aqt

        QMainWindow.__init__(main_window)
        main_window.backend = backend
        main_window.state = "startup"
        main_window.opts = opts
        main_window.col: Optional["Collection"] = None  # type: ignore

        try:  # 2.1.28+
            main_window.taskman = TaskManager(main_window)
        except TypeError:
            main_window.taskman = TaskManager()  # type: ignore

        main_window.media_syncer = MediaSyncer(main_window)

        try:  # 2.1.45+
            from aqt.flags import FlagManager

            main_window.flags = FlagManager(main_window)
        except (ImportError, ModuleNotFoundError):
            pass

        aqt.mw = main_window
        main_window.app = app
        main_window.pm = profileManager
        main_window.safeMode = False  # disable safe mode, of no use to us
        main_window.setupUI()

        post_ui_setup_callback(main_window)

        try:  # 2.1.28+
            main_window.finish_ui_setup()
        except AttributeError:
            pass

    return custom_init


@contextmanager
def patch_anki(
    post_ui_setup_callback: PostUISetupCallbackType,
) -> Iterator[str]:
    """Patch Anki to:
    - allow more fine-grained control of test execution environment
    - enable concurrent testing
    - bypass blocking update dialog
    """
    from anki.utils import checksum
    from aqt import AnkiApp, errors
    from aqt.main import AnkiQt

    old_init = AnkiQt.__init__
    old_key = AnkiApp.KEY
    old_setupAutoUpdate = AnkiQt.setupAutoUpdate
    old_maybe_check_for_addon_updates = AnkiQt.maybe_check_for_addon_updates
    old_errorHandler = errors.ErrorHandler

    patched_ankiqt_init = custom_init_factory(
        post_ui_setup_callback=post_ui_setup_callback
    )

    AnkiQt.__init__ = patched_ankiqt_init  # type: ignore
    AnkiApp.KEY = "anki" + checksum(str(uuid.uuid4()))
    AnkiQt.setupAutoUpdate = Mock()  # type: ignore[assignment]
    AnkiQt.maybe_check_for_addon_updates = Mock()  # type: ignore[assignment]
    errors.ErrorHandler = Mock()  # type: ignore[misc]

    yield AnkiApp.KEY

    AnkiQt.__init__ = old_init  # type: ignore[assignment]
    AnkiApp.KEY = old_key  # type: ignore[assignment]
    AnkiQt.setupAutoUpdate = old_setupAutoUpdate  # type: ignore[assignment]
    AnkiQt.maybe_check_for_addon_updates = (  # type: ignore[assignment]
        old_maybe_check_for_addon_updates
    )
    errors.ErrorHandler = old_errorHandler  # type: ignore[misc]


def set_qt_message_handler_installer(message_handler_installer: Callable):
    aqt.qInstallMessageHandler = message_handler_installer  # type: ignore[assignment]



# pytest-anki
#
# Copyright (C)  2019-2021 Aristotelis P. <https://glutanimate.com/>
#                and contributors (see CONTRIBUTORS file)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version, with the additions
# listed at the end of the license file that accompanied this program.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# NOTE: This program is subject to certain additional terms pursuant to
# Section 7 of the GNU Affero General Public License.  You should have
# received a copy of these additional terms immediately following the
# terms and conditions of the GNU Affero General Public License that
# accompanied this program.
#
# If not, please request a copy through one of the means of contact
# listed here: <https://glutanimate.com/contact/>.
#
# Any modifications to this file must keep this entire header intact.

import json
import socket
from contextlib import closing
from functools import reduce
from pathlib import Path
from typing import Any, Union


def create_json(path: Union[str, Path], data: dict) -> str:
    """Creates a JSON file at the specified path, preloading it with the specified data.

    Arguments:
        path {Union[str, Path]} -- path to JSON file
        data {dict} -- data to write to file

    Returns:
        Path -- path to JSON file
    """
    json_path = Path(path)
    with json_path.open("w") as f:
        f.write(json.dumps(data))
    return str(json_path)


def get_nested_attribute(obj: Any, attr: str, *args) -> Any:
    """
    Gets nested attribute from "."-separated string

    Arguments:
        obj {object} -- object to parse
        attr {string} -- attribute name, optionally including
                         "."-characters to denote different levels
                         of nesting

    Returns:
        Any -- object corresponding to attribute name

    Credits:
        https://gist.github.com/wonderbeyond/d293e7a2af1de4873f2d757edd580288
    """

    def _getattr(obj: Any, attr: str):
        return getattr(obj, attr, *args)

    return reduce(_getattr, [obj] + attr.split("."))


def find_free_port():
    # https://stackoverflow.com/a/45690594
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pytest_anki']

package_data = \
{'': ['*']}

install_requires = \
['anki>=2.1.28,<=2.1.47',
 'aqt>=2.1.28,<=2.1.47',
 'pytest-forked>=1.3.0,<2.0.0',
 'pytest-qt>=4.0.2,<5.0.0',
 'pytest-xdist>=2.3.0,<3.0.0',
 'pytest-xvfb>=2.0.0,<3.0.0',
 'pytest>=3.5.0',
 'selenium>=3.141.0,<4.0.0']

entry_points = \
{'pytest11': ['anki = pytest_anki.plugin']}

setup_kwargs = {
    'name': 'pytest-anki',
    'version': '1.0.0b3',
    'description': 'A pytest plugin for testing Anki add-ons',
    'long_description': '# pytest-anki\n\npytest-anki is a [pytest](https://docs.pytest.org/) plugin that allows developers to write tests for their [Anki add-ons](https://addon-docs.ankiweb.net/).\n\nAt its core lies the `anki_session` fixture that provides add-on authors with the ability to create and control headless Anki sessions to test their add-ons in:\n\n```python\nfrom pytest_anki import AnkiSession\n\ndef test_addon_registers_deck(anki_session: AnkiSession):\n    my_addon = anki_session.load_addon("my_addon")\n    with anki_session.load_profile()\n        with anki_session.deck_installed(deck_path) as deck_id:\n            assert deck_id in my_addon.deck_ids\n\n```\n\n`anki_session` comes with a comprehensive API that allows developers to programmatically manipulate Anki, set up and reproduce specific configurations, simulate user interactions, and much more.\n\nThe goal is to provide add-on authors with a one-stop-shop for their functional testing needs, while also enabling them to QA their add-ons against a battery of different Anki versions, catching incompatibilities as they arise.\n\n![PyPI](https://img.shields.io/pypi/v/pytest-anki) <a title="License: GNU AGPLv3" href="https://github.com/glutanimate/anki-addon-builder/blob/master/LICENSE"><img  src="https://img.shields.io/badge/license-GNU AGPLv3-f37f40.svg"></a>  <a href="https://github.com/psf/black"><img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>  [![tests](https://github.com/glutanimate/pytest-anki/actions/workflows/general.yml/badge.svg)](https://github.com/glutanimate/pytest-anki/actions/workflows/general.yml) \n\n## Disclaimer\n\n### Project State\n\n**Important**: The plugin is currently undergoing a major rewrite and expansion of its feature-set, so the documentation below is very sparse at the moment. I am working on bringing the docs up to speed, but until then, please feel free to check out the inline documentation and also take a look at the plug-in\'s tests for a number of hopefully helpful examples.\n\n### Platform Support\n\n`pytest-anki` has only been confirmed to work on Linux so far.\n\n\n## Installation\n\n### Requirements\n\n`pytest-anki` requires Python 3.8+.\n\n### Installing the latest packaged build\n\n```bash\n$ pip install pytest-anki\n```\n\nor\n\n```bash\n$ poetry add --dev pytest-anki\n```\n\n\n## Usage\n\n### Basic Use\n\nIn your tests add:\n   \n```python\nfrom pytest_anki import AnkiSession  # for type checking and completions\n\n@pytest.mark.forked\ndef test_my_addon(anki_session: AnkiSession):\n    # add some tests in here\n```\n\nThe `anki_session` fixture yields an `AnkiSession` object that gives you access to the following attributes, among others:\n\n```\napp {AnkiApp} -- Anki QApplication instance\nmw {AnkiQt} -- Anki QMainWindow instance\nuser {str} -- User profile name (e.g. "User 1")\nbase {str} -- Path to Anki base directory\n```\n\nAdditionally, the fixture provides a number of helpful methods and context managers, e.g. for initializing an Anki profile:\n\n```python\n@pytest.mark.forked\ndef test_my_addon(anki_session: AnkiSession):\n    with anki_session.profile_loaded():\n        assert anki_session.collection\n```\n\n\n### Configuring the Anki Session\n\nYou can customize the Anki session context by passing arguments to the `anki_session` fixture using pytest\'s indirect parametrization, e.g.\n\n```python\nimport pytest\n\n@pytest.mark.forked\n@pytest.mark.parametrize("anki_session", [dict(load_profile=True)], indirect=True)\ndef test_my_addon(anki_session: AnkiSession):\n    # profile / collection already pre-loaded!\n    assert anki_session.collection\n```\n\n## Additional Notes\n\n### When to use pytest-anki\n\nRunning your test in an Anki environment is expensive and introduces an additional layer of confounding factors. If you can `mock` your Anki runtime dependencies away, then that should always be your first tool of choice.\n\nWhere `anki_session` comes in handy is further towards the upper levels of the test pyramid, i.e. functional tests, end-to-end tests, and UI tests. Additionally the plugin can provide you with a convenient way to automate testing for incompatibilities with Anki and other add-ons.\n\n### The importance of forking your tests\n\nYou might have noticed that most of the examples above use a `@pytest.mark.forked` decorator. This is because, while the plugin does attempt to tear down Anki sessions as cleanly as possible on exit, this process is never quite perfect, especially for add-ons that monkey-patch Anki.\n\nWith unforked test runs, factors like that can lead to unexpected behavior, or worse still, your tests crashing. Forking a new subprocess for each test bypasses these limitations, and therefore my advice would be to mark any `anki_session` tests as forked by default.\n\nTo do this in batch for an entire test module, you can use the following pytest hook:\n\n```python\ndef pytest_collection_modifyitems(items):\n    for item in items:\n        item.add_marker("forked")\n```\n\nFuture versions of `pytest-anki` will possibly do this by default.\n\n### Automated Testing\n\n`pytest-anki` is designed to work well with continuous integration systems such as GitHub actions. For an example see `pytest-anki`\'s own [GitHub workflows](./.github/workflows/).\n\n\n### Troubleshooting\n\n#### pytest hanging when using xvfb\n\nEspecially if you run your tests headlessly with `xvfb`, you might run into cases where pytest will sometimes appear to hang. Oftentimes this is due to blocking non-dismissable prompts that your add-on code might invoke in some scenarios. If you suspect that might be the case, my advice would be to temporarily bypass `xvfb` locally via `pytest --no-xvfb` to show the UI and manually debug the issue.\n\n## Contributing\n\nContributions are welcome! To set up `pytest-anki` for development, please first make sure you have Python 3.8+ and [poetry](https://python-poetry.org/docs/) installed, then run the following steps:\n\n```\n$ git clone https://github.com/glutanimate/pytest-anki.git\n\n$ cd pytest-anki\n\n# Either set up a new Python virtual environment at this stage\n# (e.g. using pyenv), or let poetry create the venv for you\n\n$ make install\n```\n\nBefore submitting any changes, please make sure that `pytest-anki`\'s checks and tests pass:\n\n```bash\nmake check\nmake lint\nmake test\n```\n\nThis project uses `black`, `isort` and `autoflake` to enforce a consistent code style. To auto-format your code you can use:\n\n```bash\nmake format\n```\n\n## License and Credits\n\n*pytest-anki* is\n\n*Copyright © 2019-2021 [Aristotelis P.](https://glutanimate.com/contact/) (Glutanimate) and [contributors](./CONTRIBUTORS)*\n\n*Copyright © 2017-2019 [Michal Krassowski](https://github.com/krassowski/anki_testing)*\n\n*Copyright © 2017-2021 [Ankitects Pty Ltd and contributors](https://github.com/ankitects/)*\n\n\nAll credits for the original idea for creating a context manager to test Anki add-ons with go to Michal. _pytest-anki_ would not exist without his [anki_testing](https://github.com/krassowski/anki_testing) project.\n\nI would also like to extend a heartfelt thanks to [AMBOSS](https://github.com/amboss-mededu/) for their major part in supporting the development of this plugin! Most of the recent feature additions leading up to v1.0.0 of the plugin were implemented as part of my work on the [AMBOSS add-on](https://www.amboss.com/us/anki-amboss).\n\n_pytest-anki_ is free and open-source software. Its source-code is released under the GNU AGPLv3 license, extended by a number of additional terms. For more information please see the [license file](https://github.com/glutanimate/pytest-anki/blob/master/LICENSE) that accompanies this program.\n\nThis program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY. Please see the license file for more details.',
    'author': 'Aristotelis P. (Glutanimate)',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/glutanimate/pytest-anki',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# pytest-anki
#
# Copyright (C)  2017-2021 Ankitects Pty Ltd and contributors
# Copyright (C)  2017-2019 Michal Krassowski <https://github.com/krassowski>
# Copyright (C)  2019-2021 Aristotelis P. <https://glutanimate.com/>
#                and contributors (see CONTRIBUTORS file)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version, with the additions
# listed at the end of the license file that accompanied this program.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# NOTE: This program is subject to certain additional terms pursuant to
# Section 7 of the GNU Affero General Public License.  You should have
# received a copy of these additional terms immediately following the
# terms and conditions of the GNU Affero General Public License that
# accompanied this program.
#
# If not, please request a copy through one of the means of contact
# listed here: <https://glutanimate.com/contact/>.
#
# Any modifications to this file must keep this entire header intact.

import os
import shutil
import tempfile
from contextlib import contextmanager, nullcontext
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional, Tuple
from unittest import mock

from PyQt5.QtCore import qInstallMessageHandler

from ._anki import AnkiStateUpdate, update_anki_colconf_state, update_anki_profile_state
from ._errors import AnkiSessionError
from ._patch import (
    patch_anki,
    post_ui_setup_callback_factory,
    set_qt_message_handler_installer,
)
from ._qt import QtMessageMatcher
from ._session import AnkiSession
from ._types import PathLike
from ._util import find_free_port

if TYPE_CHECKING:
    from pytestqt.qtbot import QtBot


QTWEBENGINE_REMOTE_DEBUGGING = "QTWEBENGINE_REMOTE_DEBUGGING"


@contextmanager
def temporary_user(anki_base_dir: str, name: str, lang: str) -> Iterator[str]:

    from aqt.profiles import ProfileManager

    pm = ProfileManager(base=anki_base_dir)

    pm.setupMeta()
    pm.setLang(lang)
    pm.create(name)

    pm.name = name

    yield name

    if pm.db:
        # reimplement pm.remove() to avoid trouble with trash
        profile_folder = pm.profileFolder()
        if os.path.exists(profile_folder):
            shutil.rmtree(profile_folder, ignore_errors=True)
        pm.db.execute("delete from profiles where name = ?", name)
        pm.db.commit()


@contextmanager
def base_directory(base_path: str, base_name: str) -> Iterator[str]:
    if not os.path.isdir(base_path):
        os.mkdir(base_path)
    anki_base_dir = tempfile.mkdtemp(prefix=f"{base_name}_", dir=base_path)
    yield anki_base_dir
    shutil.rmtree(anki_base_dir, ignore_errors=True)


@contextmanager
def anki_running(
    qtbot: "QtBot",
    base_path: str = tempfile.gettempdir(),
    base_name: str = "anki_base",
    profile_name: str = "User 1",
    lang: str = "en_US",
    load_profile: bool = False,
    preset_anki_state: Optional[AnkiStateUpdate] = None,
    packed_addons: Optional[List[PathLike]] = None,
    unpacked_addons: Optional[List[Tuple[str, PathLike]]] = None,
    addon_configs: Optional[List[Tuple[str, Dict[str, Any]]]] = None,
    enable_web_debugging: bool = True,
) -> Iterator[AnkiSession]:
    """Context manager that safely launches an Anki session, cleaning up after itself

    Keyword Arguments:
        base_path {str} -- Path to write Anki base folder to
            (default: system-wide temporary directory)

        base_name {str} -- Base folder name (default: {"anki_base"})

        profile_name {str} -- User profile name (default: {"User 1"})

        lang {str} -- Language to use for the user profile (default: {"en_US"})

        load_profile {bool} -- Whether to preload Anki user profile (with collection)
            (default: {False})

        preset_anki_state {Optional[pytest_anki.AnkiStateUpdate]}:
            Allows pre-configuring Anki object state, as described by a PresetAnkiState
            dataclass. This includes the three main configuration storages used by
            add-ons, mw.col.conf (colconf_storage), mw.pm.profile (profile_storage),
            and mw.pm.meta (meta_storage).

            The provided data is applied on top of the existing data in each case
            (i.e. in the same way as dict.update(new_data) would).

            State specified in this manner is guaranteed to be pre-configured ahead of
            add-on load time (in the case of meta_storage), or ahead of
            gui_hooks.profile_did_open fire time (in the case of colconf_storage and
            profile_storage).

            Please note that, in the case of colconf_storage and profile_storage, the
            caller is responsible for either passing 'load_profile=True', or manually
            loading the profile at a later stage.

        packed_addons {Optional[List[PathLike]]}: List of paths to .ankiaddon-packaged
            add-ons that should be installed ahead of starting Anki

        unpacked_addons {Optional[List[Tuple[str, PathLike]]]}:
            List of unpacked add-ons that should be installed ahead of starting Anki.
            Add-ons need to be specified as tuple of the add-on package name under which
            to install the add-on, and the path to the source folder (the package
            folder containing the add-on __init__.py)

        addon_configs {Optional[List[Tuple[str, Dict[str, Any]]]]}:
            List of add-on package names and config values to set the user configuration
            for the specified add-on to. Useful for simulating specific config set-ups.
            Each list member needs to be specified as a tuple of add-on package name
            and dictionary of user configuration values to set.

        web_debugging_port {Optional[int]}:
            If specified, launches Anki with QTWEBENGINE_REMOTE_DEBUGGING set, allowing
            you to remotely debug Qt web engine views.

    Returns:
        Iterator[AnkiSession] -- [description]

    Yields:
        Iterator[AnkiSession] -- [description]
    """

    import aqt
    from aqt import gui_hooks

    with base_directory(base_path=base_path, base_name=base_name) as anki_base_dir:

        # Callback to run between main UI initialization and finishing steps of UI
        # initialization (add-on loading time)

        post_ui_setup_callback = post_ui_setup_callback_factory(
            anki_base_dir=anki_base_dir,
            packed_addons=packed_addons,
            unpacked_addons=unpacked_addons,
            addon_configs=addon_configs,
            preset_anki_state=preset_anki_state,
        )

        # Apply preset Anki profile and collection.conf storage on profile load

        def profile_loaded_callback():
            if not aqt.mw or not preset_anki_state:
                return
            update_anki_profile_state(
                main_window=aqt.mw, anki_state_update=preset_anki_state
            )
            update_anki_colconf_state(
                main_window=aqt.mw, anki_state_update=preset_anki_state
            )

        if preset_anki_state and (
            preset_anki_state.colconf_storage or preset_anki_state.profile_storage
        ):
            gui_hooks.profile_did_open.append(profile_loaded_callback)
            profile_hooked = True
        else:
            profile_hooked = False

        # Start Anki session

        with patch_anki(post_ui_setup_callback=post_ui_setup_callback):
            with temporary_user(
                anki_base_dir=anki_base_dir, name=profile_name, lang=lang
            ) as user_name:

                environment = {}

                if enable_web_debugging:
                    web_debugging_port = find_free_port()
                    if web_debugging_port is None:
                        raise OSError("Could not find a free port for remote debugging")
                    environment[QTWEBENGINE_REMOTE_DEBUGGING] = str(web_debugging_port)
                else:
                    web_debugging_port = None

                with mock.patch.dict(os.environ, environment):

                    if os.environ.get(QTWEBENGINE_REMOTE_DEBUGGING):

                        # We want to wait until remote debugging started to yield the
                        # Anki session, so we monitor Qt's log for the corresponding msg
                        qt_message_matcher = QtMessageMatcher(
                            "Remote debugging server started successfully"
                        )

                        # On macOS, Anki does not install a custom message
                        # handler, so we can install our own directly:
                        qInstallMessageHandler(qt_message_matcher)

                        # On Windows and Linux, we need to monkey-patch the
                        # message handler installer to make sure that ours is
                        # not switched out when aqt runs
                        def install_message_handler(message_handler):
                            def message_handler_wrapper(*args, **kwargs):
                                qt_message_matcher(*args, **kwargs)
                                return message_handler(*args, **kwargs)

                            qInstallMessageHandler(message_handler_wrapper)

                        set_qt_message_handler_installer(install_message_handler)

                        maybe_wait_for_web_debugging = qtbot.wait_signal(
                            qt_message_matcher.match_found
                        )
                    else:
                        maybe_wait_for_web_debugging = nullcontext()

                    with maybe_wait_for_web_debugging:
                        # We don't pass in -p <profile> in order to avoid
                        # profileloading. This helps replicate the profile
                        # availability at add-on init time for most users. Anki
                        # will automatically open the profile at mw.setupProfile
                        # time in single-profile setups
                        app = aqt._run(argv=["anki", "-b", anki_base_dir], exec=False)

                    mw = aqt.mw

                    if mw is None or app is None:
                        raise AnkiSessionError("Main window not initialized correctly")

                    anki_session = AnkiSession(
                        app=app,
                        mw=mw,
                        user=user_name,
                        base=anki_base_dir,
                        qtbot=qtbot,
                        web_debugging_port=web_debugging_port,
                    )

                    if not load_profile:
                        yield anki_session

                    else:
                        with anki_session.profile_loaded():
                            yield anki_session

                    # Undo monkey-patch if applied
                    set_qt_message_handler_installer(qInstallMessageHandler)

    # NOTE: clean up does not seem to work properly in all cases,
    # so use pytest-forked for now

    # clean up what was spoiled
    if aqt.mw:
        aqt.mw.cleanupAndExit()

    # remove hooks added by pytest-anki

    if profile_hooked:
        gui_hooks.profile_did_open.remove(profile_loaded_callback)

    # remove hooks added during app initialization
    from anki import hooks

    hooks._hooks = {}

    # test_nextIvl will fail on some systems if the locales are not restored
    import locale

    locale.setlocale(locale.LC_ALL, locale.getdefaultlocale())  # type: ignore
